//Ethan Alexander Shulman 2022 "Color Space" April 13, 2022

//scene rendering options
#define VIEW_STEPS 60
#define VIEW_STEP_DITHER .9
#define VIEW_EPSILON 1e-1
//#define VIEW_EPSILON_MUL 1.14
#define LIGHTING_STEPS 0
#define LIGHTING_STEP_DITHER .1
#define LIGHTING_EPSILON 1e-2
//#define COLOR_MAX 100.
#define DENOISE_RANGE 0
#define DENOISE_NORMAL_THRESHOLD .99
#define DENOISE_DEPTH_THRESHOLD 1e1
#define LOW_DISCREPANCY 1
//#define ANIMATION_TIME 60.
#define ANIMATION_SAMPLES 1000
#ifdef ANIMATION_TIME
out float time;
#else
float time = 0.;
#endif

//constants
#define PI 3.14159265



float twave(float x) {
	return 1.-abs(fract(x)-.5)*2.;
}

mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))



/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	vec4 fhs = vec4(hash(uvec4(frame)))*I2F;
	return mat3(0,0,-2000+fhs.x*100.,//time*400.,
			0,0,0,
			1,0,0);//mix(.25,4.,twave(time/60.)),0,0);
}


//distance dithering for volumetrics
#define START_VOLUME(posSeed) vec2 opacityRand = fract(baseRand.zw+128.*dot(abs(posSeed),baseRand.zxy)*baseRand.xy)
#define VOLUME(dst,skip,alpha,anisotropy) max(dst,-1e3+(skip+1e3)*step(alpha+nsampling*anisotropy,opacityRand.x))

//dither materials for chromatic separation
#define START_CHROMA(posSeed) float chromaFrame = mod(baseRand.x+156.*dot(abs(posSeed),baseRand.zwy),3.)
#define CHROMA(mt,rc,gc,bc) if (chromaFrame < 1.) {mt[0]*=vec3(3,0,0);rc;}\
else if (chromaFrame < 2.) {mt[0]*=vec3(0,3,0);gc;}\
else {mt[0]*=vec3(0,0,3);bc;}



//functions defining scene
//returns vec2(distance, materialId) of nearest surface
vec3 lhp;
vec4 chv;
vec2 geometry(vec3 p) {
	vec3 fp = floor(p/100.+10000.),
		rp = mod(abs(p),100.)-50.;
	if (lhp != fp) chv = vec4(hash(uvec4(fp,0)))*I2F;
	return vec2(length(rp-(chv.xyz-.5)*99.)-chv.w*1.,0.);
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;
float refractionIndex;*/
mat3 material(vec3 p, vec3 n, float mat) {
	vec3 fp = floor(p/100.+10000.);
	return mat3(pow(vec3(hash(uvec4(fp,1)))*I2F,vec3(3)), 0,4,1, 1,0,0);
}

//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	return vec3(0);
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return background(p,d);
}

